/* Copyright (c) 2017, Haakan T. Johansson */
/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the authors nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define TEST_STOREMASK      ((1 << (DPTC_BITS))-1)

#define COLLDIFF 1

#define DPTC_GET_TOTALBITS 1

#define ADS_MAX_LEVEL 4

#include "dptc_pack.c"
#include "dptc_unpack.c"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define BUFLEN 100000

uint16_t vector16[BUFLEN];
uint32_t vector32[BUFLEN];
uint32_t packed16[BUFLEN];
uint32_t packed32[BUFLEN];
uint16_t unpack16[BUFLEN];
uint32_t unpack32[BUFLEN];

uint32_t ads_vector32[BUFLEN];
uint32_t ads_packed32[BUFLEN];
uint32_t ads_unpack32[BUFLEN];
int n = 0;

#if COLLDIFF
#define NUMCOLLDIFF (0x10000*4)

/* Plus one to have a zero dummy entry at the end (for summing). */
uint16_t colldiff[NUMCOLLDIFF+1];
#endif


typedef struct statistics_t
{
  int traces;
  uint64_t totalpoints;
  uint64_t packedbytes;
  uint64_t packedbits;

  uint64_t part_totalpoints[9];
  uint64_t part_packedbytes[9];

  uint64_t reducedpoints;
  uint64_t reducedpacked;
  uint64_t reducedpackedbits;
  
} statistics;

statistics _stat;

int diffs[0x1000];
int diffs2[0x1000];

uint32_t max_ampl = 0;
double sum_log_ampl = 0.0;
double sum_log_ampl_n = 0.0;

uint16_t vector16_diff[BUFLEN];

FILE *binfile  = NULL;
FILE *difffile = NULL;

uint64_t xorshift(uint64_t x)
{
  x ^= x >> 13;
  x ^= x << 7;
  x ^= x >> 17;
  return x;
}

void test_pack()
{
  int i;
  size_t np16, np32, ads_np32;
  int totalbits16, totalbits32, ads_totalbits32;
  uint64_t xorseed;
  uint32_t level;
  
  if (!n)
    {
#if DEBUGPACK || DPTC_DEBUG_PACK
      /* VHDL testbench will write a marker also for the first Event mark. */
      printf ("OUTEND\n");
#endif
      return;
    }

  uint32_t minv = 0xffffffff;
  uint32_t maxv = 0;
  uint32_t ampl;

  for (i = 0; i < n; i++)
    {
      if (vector32[i] < minv) minv = vector32[i];
      if (vector32[i] > maxv) maxv = vector32[i];
    }

  ampl = maxv - minv;

  if (ampl > max_ampl)
    max_ampl = ampl;

  if (ampl > 0)
    {
      sum_log_ampl += log(ampl);
      sum_log_ampl_n++;
    }

#if 0
  if (maxv - minv < 100)
    {
      n = 0;
      return;
    }
#endif

#if COLLDIFF
  for (i = 4; i < n-4; i += 5)
    {
      int32_t avg =
	(vector32[i-4] + vector32[i-3] + vector32[i-2] + vector32[i-1] +
	 vector32[i+1] + vector32[i+2] + vector32[i+3] + vector32[i+4]);
      int32_t diff = ((int32_t) vector32[i]) * 4 - avg / 2;
      /*
      int32_t avg =
	(vector32[i-2] + vector32[i-1] +
	 vector32[i+1] + vector32[i+2]);
      int32_t diff = ((int32_t) vector32[i]) * 4 - avg;
      */
      colldiff[diff & (NUMCOLLDIFF-1)]++;
    }
#endif

  fwrite(vector16,sizeof(vector16[0]),n,binfile);

  for (i = 0; i < n-1; i++)
    {
      vector16_diff[i] = vector16[i+1] - vector16[i];
    }

  fwrite(vector16_diff,sizeof(vector16_diff[0]),n-1,difffile);

  memset(packed16, 0x55, sizeof (packed16));

  np16 = dptc_compress16(vector16, n,
			 packed16, n,
			 DPTC_BITS < 16 ? DPTC_BITS : 16,
			 &totalbits16);

  np32 = dptc_compress32(vector32, n,
			 packed32, n,
			 DPTC_BITS,
			 &totalbits32);

  if (DPTC_BITS <= 16)
    {
      if (np16 != np32)
	{
	  printf ("dptc_compress16 and dptc_compress32 give different "
		  "lengths.");
	  exit(1);
	}

      for (i = 0; i < np32; i++)
	if (packed16[i] != packed32[i])
	  {
	    printf ("16-32 %4d: %5d != %5d\n", i, packed16[i], packed32[i]);
	    exit(1);
	  }
    }

#if DEBUGPACK || DPTC_DEBUG_PACK
  for (i = 0; i < np32; i++)
    {
      printf ("OUTPUT: 0x%08X = ", packed32[i]);
#ifdef DPTC
      dptc_printbits(packed32[i],32);
#else
      printbits(packed16[i],32);
#endif
      printf ("\n");
    }
  printf ("OUTEND\n");
#endif

  /* Create some random level marks for the ads_vector32. */

  xorseed = xorshift(ads_vector32[0] +
		     (((uint64_t) ads_vector32[n-1]) << 32));

  level = xorseed % (ADS_MAX_LEVEL + 1);

  ads_vector32[0] |= level << 28;

  for (i = 1; i < n; i++)
    {
      xorseed = xorshift(xorseed ^ ads_vector32[i]);

      switch (xorseed & 3)
	{
	case 0:
	  if (level < ADS_MAX_LEVEL)
	    level++;
	  break;
	case 1:
	  if (level > 0)
	    level--;
	  break;
	case 2:
	case 3:
	  /* no change */
	  break;
	}

      ads_vector32[i] |= level << 28;
    }

  ads_np32 = dptc_ads_compress32(ads_vector32, n,
				 ads_packed32, n,
				 DPTC_BITS + ADS_MAX_LEVEL, ADS_MAX_LEVEL,
				 &ads_totalbits32);

#if DEBUGPACK || DPTC_DEBUG_PACK
  for (i = 0; i < ads_np32; i++)
    {
      printf ("ADSOUTPUT: 0x%08X = ", ads_packed32[i]);
      dptc_printbits(ads_packed32[i],32);
      printf ("\n");
    }
  printf ("ADSOUTEND\n");
#endif

  unpack16[n] = 0xdead;
  unpack32[n] = 0xdeadbeef;
  ads_unpack32[n] = 0xdeadbeef;

  int ret = dptc_unpack16(packed16, np16,
			  unpack16, n,
			  DPTC_BITS < 16 ? DPTC_BITS : 16);

  if (ret != 0)
    {
      printf ("dptc_unpack16 returned failure\n");
      exit(1);
    }

  ret = dptc_unpack32(packed32, np32,
		      unpack32, n,
		      DPTC_BITS);

  if (ret != 0)
    {
      printf ("dptc_unpack32 returned failure\n");
      exit(1);
    }

  ret = dptc_ads_unpack32(ads_packed32, ads_np32,
			  ads_unpack32, n,
			  DPTC_BITS + ADS_MAX_LEVEL, ADS_MAX_LEVEL);

  if (ret != 0)
    {
      printf ("dptc_unpack32 returned failure\n");
      exit(1);
    }

  int ok = 1;

  for (i = 0; i < n; i++)
    if (vector16[i] != unpack16[i])
      {
	printf ("%4d: %5d (%04x) != %5d (%04x)\n",
		i, vector16[i], vector16[i], unpack16[i], unpack16[i]);
	ok = 0;
      }

  if (unpack16[n] != 0xdead)
    {
      printf ("%4d: 0x%08x != 0xdeadbeef - buffer overrun\n",
	      i, unpack16[i]);
      ok = 0;
    }

  for (i = 0; i < n; i++)
    if (vector32[i] != unpack32[i])
      {
	printf ("%4d: %5d (%04x) != %5d (%04x)\n",
		i, vector32[i], vector32[i], unpack32[i], unpack32[i]);
	ok = 0;
      }

  if (unpack32[n] != 0xdeadbeef)
    {
      printf ("%4d: 0x%08x != 0xdeadbeef - buffer overrun\n",
	      i, unpack32[i]);
      ok = 0;
    }

  for (i = 0; i < n; i++)
    if (ads_vector32[i] != ads_unpack32[i])
      {
	printf ("%4d: %5d (%04x) != %5d (%04x)\n",
		i,
		ads_vector32[i], ads_vector32[i],
		ads_unpack32[i], ads_unpack32[i]);
	ok = 0;
      }

  if (ads_unpack32[n] != 0xdeadbeef)
    {
      printf ("%4d: 0x%08x != 0xdeadbeef - buffer overrun\n",
	      i, ads_unpack32[i]);
      ok = 0;
    }

  if (!ok)
    {
      printf ("mismatch\n");
      exit(1);
    }

  _stat.traces++;
  _stat.totalpoints += n;
  _stat.packedbytes += np32 * sizeof (packed16[0]);
  _stat.packedbits += totalbits32;

  /* printf ("%5d %5d %d\n", minv, maxv, np * sizeof (packed16[0])); */

#if !DEBUGPACK && !DPTC_DEBUG_PACK
  /* Trim the trace to avoid noise in the beginning and end.
   * Will be less to pack, but harder... 
   */

  {
    uint16_t *start = vector16;
    uint16_t *end = vector16 + n;

#define REDUCE_NOISE_TOL 6

    while (end > start + 4)
      {
	int d01 = abs((int) start[0]) - ((int) start[1]);
	int d02 = abs((int) start[0]) - ((int) start[2]);
	int d12 = abs((int) start[1]) - ((int) start[2]);

	if (d01 > REDUCE_NOISE_TOL ||
	    d02 > REDUCE_NOISE_TOL ||
	    d12 > REDUCE_NOISE_TOL)
	  break;
	start++;
      }
    while (end > start + 4)
      {
	int d01 = abs((int) end[-1]) - ((int) end[-2]);
	int d02 = abs((int) end[-1]) - ((int) end[-3]);
	int d12 = abs((int) end[-2]) - ((int) end[-3]);

	if (d01 > REDUCE_NOISE_TOL ||
	    d02 > REDUCE_NOISE_TOL ||
	    d12 > REDUCE_NOISE_TOL)
	  break;
	end--;
      }

    np16 = dptc_compress(start, end-start,
			 packed16, n,
			 DPTC_BITS,
			 &totalbits16);
    
    _stat.reducedpoints += end-start;
    _stat.reducedpacked += np16 * sizeof (packed16[0]);
    _stat.reducedpackedbits += totalbits16;
  }

#if 0
  for (i = 0; i < 9; i++)
    {
      uint16_t *start = vector;
      uint16_t *end = vector + n;

      switch (i)
	{
	case 0:
	  start = vector;
	  end = vector + 10;
	  break;
	case 1:
	  start = vector;
	  end = vector + 20;
	  break;
	case 2:
	  start = vector;
	  end = vector + 30;
	  break;
	case 3:
	  start = vector + 30;
	  end = vector + 53;
	  break;
	case 4:
	  start = vector + 30;
	  end = vector + 76;
	  break;
	case 5:
	  start = vector + 30;
	  end = vector + 100;
	  break;
	case 6:
	  start = vector + 100;
	  end = vector + 133;
	  break;
	case 7:
	  start = vector + 100;
	  end = vector + 166;
	  break;
	case 8:
	  start = vector + 100;
	  end = vector + 200;
	  break;
	}      

      np16 = dptc_compress(start, end-start,
			   packed16, n,
			   &totalbits);
      
      _stat.part_totalpoints[i] += end-start;
      _stat.part_packedbytes[i] += np16 * sizeof (packed[0]);
    }
#endif

#endif

#if 0
  int dump = 0;

  int d3 = 0;
  int d2 = 0;
  int d1 = 0;

  for (i = 1; i < n; i++)
    {
      int diff = (vector16[i] - vector16[i-1]) & 0xfff;
      
      diffs[diff]++;

      if (diff >= 30 && diff < 4096 - 30)
	dump = 1;

      diff = (((int) vector16[i]) - (int) vector16[i-1]);

      int sd = diff;

      if ((d3 < 0 && d2 < 0 && d1 < 0) ||
	  (d3 > 0 && d2 > 0 && d1 > 0))
	sd = diff - d1;
      else
	sd = diff + (d1 + d2) / 2;

      diffs2[sd & 0xfff]++;
      
      d3 = d2;
      d2 = d1;
      d1 = diff;
    }

  dump = 0;

  if (maxv - minv > 100)
    dump = 1;

  if (dump)
    {
      for (i = 0; i < n; i++)
	printf ("DUMP %5d\n",vector16[i]);
      /* printf ("\n"); */
    }
#endif

  n = 0;
}

int main()
{
  int i;
  char *s;

  binfile = fopen("traces.bin","w");
  difffile = fopen("traces_diff.bin","w");

  memset(&_stat, 0, sizeof (_stat));
  memset(diffs, 0, sizeof (diffs));
#if COLLDIFF
  memset(colldiff, 0, sizeof (colldiff));
#endif

  while (!feof(stdin))
    {
      char buf[1024];

      s = fgets(buf, sizeof(buf), stdin);

      if (s == NULL)
	break;

      if (strstr(buf, "Event"))
	test_pack();
      else
	{
	  int v;
	  int r;
	  
	  r = sscanf(buf, " %d ", &v);

	  if (r == 1)
	    {
	      vector32[n] = v & TEST_STOREMASK;
	      vector16[n] = (uint16_t) vector32[n];
	      ads_vector32[n] = vector32[n];
	      n++;
	    }
	}
    }

  printf ("Vector with %d items read.\n", n);
  
  test_pack();

#if 0

  for (i = 2048; i < 4096; i++)
    printf (" %d", diffs[i]);
  for (i = 0; i < 2048; i++)
    printf (" %d", diffs[i]);
  printf ("\n");

  for (i = 0; i < 4096; i++)
    printf (" %d", diffs2[i]);
  printf ("\n");
#endif

#if COLLDIFF
  uint64_t sum = 0;
  uint64_t maxcolldiff = 0;
    
  for (i = 0; i < NUMCOLLDIFF; i++)
    {
      sum += colldiff[i];
      if (colldiff[i] > maxcolldiff)
	maxcolldiff = colldiff[i];
    }

  uint64_t sum2 = 0;
  int end2i = 0;
  
  for (i = 0; i < NUMCOLLDIFF/2; i++)
    {
      sum2 += colldiff[i] + colldiff[NUMCOLLDIFF-i];

      if (sum2 > sum / 2)
	{
	  end2i = i;
	  break;
	}
    }

  if (!end2i)
    end2i = 1;

#define WINDOW 5

  double sumoutn = 0;
  double nbin = 0;

  printf ("end2i: %d\n",end2i);
  
  for (i = end2i*WINDOW; i <= end2i*2*WINDOW && i < NUMCOLLDIFF / 2; i++)
    {
      sumoutn += colldiff[i] + colldiff[NUMCOLLDIFF-i];
      nbin += 2;
    }

  double avgbkgnd = sumoutn / nbin;
  
  double sumn = 0;
  double sumx = 0;
  double sumxx = 0;

  for (i = 0; i <= end2i*WINDOW && i < NUMCOLLDIFF / 2; i++)
    {
      sumn  += (colldiff[i]-avgbkgnd);
      sumx  += (colldiff[i]-avgbkgnd) * i * 0.25;
      sumxx += (colldiff[i]-avgbkgnd) * (i * 0.25) * (i * 0.25);

      sumn  += (colldiff[NUMCOLLDIFF-i]-avgbkgnd);
      sumx  += (colldiff[NUMCOLLDIFF-i]-avgbkgnd) * i * -0.25;
      sumxx += (colldiff[NUMCOLLDIFF-i]-avgbkgnd) * (i * 0.25) * (i * 0.25);
    }

  double mean = sumx / sumn;
  double std = sqrt((sumxx - sumx * sumx / sumn) / (sumn - 1));

  /* sqrt(9./8.) is correction factor for surrounding average of 8 */
  std = std / sqrt(9./8.);
#endif

  printf ("Traces: %6d  Samples/trace: %6.1f  Bits/sample: %5.2f"
#if COLLDIFF
	  /*
	  " :  %8d %8d %5d %5d  %8.1f %8.1f"
	  */

	  "  %6.3f %6.3f %5.2f  %5d  %6.1f"
#endif
	  " (%6.1f : %5.2f)\n",
	  _stat.traces,
	  ((double) _stat.totalpoints) / _stat.traces,
	  (1.0*_stat.packedbits/*bytes * 8.0*/) / _stat.totalpoints,
#if COLLDIFF
	  /*
	  (int) sum, (int) sum2, (int) maxcolldiff, (int) i, sumn, avgbkgnd,
	  */
	  mean, std,
	  /*1./8.*log(1.0 + pow(std*pow(2.0, 2.95-1.5),8)) / log(2) + 1.5,*/
	  /*1./8.*log(1.0 + pow(std,8)) / log(2) + 2.95,*/
	  log(std) / log(2) + 2.95,
#endif
	  max_ampl,
	  exp(sum_log_ampl / sum_log_ampl_n),
	  ((double) _stat.reducedpoints) / _stat.traces,
	  (1.0*_stat.reducedpackedbits/* * 8.0*/) / _stat.reducedpoints);

  for (i = 0; i < 3; i++)
    {
      printf ("%5.2f %5.2f %5.2f\n",
	      (_stat.part_packedbytes[i*3+0] * 8.0) / _stat.part_totalpoints[i*3+0],
	      (_stat.part_packedbytes[i*3+1] * 8.0) / _stat.part_totalpoints[i*3+1],
	      (_stat.part_packedbytes[i*3+2] * 8.0) / _stat.part_totalpoints[i*3+2]);
      printf ("%5.2f %5.2f %5.2f\n",
	      (_stat.part_packedbytes[i*3+0] * 8.0) / 100,
	      (_stat.part_packedbytes[i*3+1] * 8.0) / 100,
	      (_stat.part_packedbytes[i*3+2] * 8.0) / 100);
    }

  fclose(binfile);
  fclose(difffile);

  return 0;
}
