/* Copyright (c) 2017, Haakan T. Johansson */
/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the authors nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define TEST_STOREMASK      ((1 << (DPTC_BITS))-1)

#include <stdio.h>
#include "dptc_pack.c"
#include "dptc_unpack.c"
#include <float.h>
#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>

#define ADS_MAX_LEVEL 4

#define BUFLEN    8192
#define TRACES    2000

size_t len[TRACES];
uint16_t vector16[TRACES][BUFLEN];
uint32_t vector32[TRACES][BUFLEN];
uint32_t ads_vector32[TRACES][BUFLEN];
uint32_t packed16[TRACES][BUFLEN];
uint32_t packed32[TRACES][BUFLEN];
uint32_t ads_packed32[TRACES][BUFLEN];
size_t np16[TRACES];
size_t np32[TRACES];
size_t ads_np32[TRACES];
uint16_t unpack16[TRACES][BUFLEN];
uint32_t unpack32[TRACES][BUFLEN];
uint32_t ads_unpack32[TRACES][BUFLEN];
size_t samples;

/* This has almost no effect at all - we are not memory bandwidth
 * limited.  So disabling.
 */
#define TESTSPEED_UNPACK_SAME_STRUCT  0

#if TESTSPEED_UNPACK_SAME_STRUCT
#define unp_k  0
#else
#define unp_k  k
#endif

char trash_vector[16*1024*1024];

int readtraces()
{
  int k = 0;
  int n;
  char *s;

  samples = 0;
  len[k] = 0;
  
  while (!feof(stdin))
    {
      char buf[1024];

      s = fgets(buf, sizeof(buf), stdin);

      if (s == NULL)
	break;

      if (strstr(buf, "Event"))
	{
	  k++;
	  len[k] = 0;

  	  if (k >= TRACES)
	    break;
	}
      else
	{
	  int v;
	  int r;
	  
	  r = sscanf(buf, " %d ", &v);

	  if (r == 1)
	    {
	      vector32[k][len[k]] = v & TEST_STOREMASK;
	      vector16[k][len[k]] = (uint16_t) vector32[k][len[k]];
	      ads_vector32[k][len[k]] = vector32[k][len[k]];
	      len[k]++;
	      samples++;
	    }
	}
    }

  return k;
}

int trash_cache()
{
  int sum;
  int k;

  for (k = 0; k < (int) sizeof (trash_vector); k++)
    sum += trash_vector[k]++;

  return sum;
}

struct speed_measure
{
  struct timeval _a, _b;
  int    _rounds;
  double _elapsed;
  int    _sum;
};

void init_speed(struct speed_measure *m)
{
  m->_rounds  = 0;
  m->_elapsed = DBL_MAX;
}

int check_speed(struct speed_measure *m, double *final)
{
  double elapsed;
  double improve;

  if (m->_rounds)
    {
      gettimeofday(&m->_b, NULL); /* First in function! */

      elapsed =    (m->_b.tv_sec  - m->_a.tv_sec) +
	0.000001 * (m->_b.tv_usec - m->_a.tv_usec);

      improve = (m->_elapsed - elapsed) / m->_elapsed;

      if (elapsed < m->_elapsed)
	m->_elapsed = elapsed;

      /*
      if (m->_rounds > 1)
	printf ("%d %.6f\n", m->_rounds, m->_elapsed);
      */

      if (m->_rounds > 10 && /* At least 5 attempts. */
	  improve < 0.0001)
	{
	  *final = m->_elapsed;
	  return 0;
	}
    }

  m->_rounds++;

  m->_sum += trash_cache();

  gettimeofday(&m->_a, NULL); /* Last before return! */

  return 1;
}

int main()
{
  int i;
  int k, l;
  struct timeval a, b;
  double elapsed;
  double elapsed_compress16 = DBL_MAX, elapsed_unpack16 = DBL_MAX;
  double elapsed_compress32 = DBL_MAX, elapsed_unpack32 = DBL_MAX;
  double elapsed_ads_compress32 = DBL_MAX, elapsed_ads_unpack32 = DBL_MAX;
  int traces;
  struct speed_measure m;

  for (k = 0; k < (int) sizeof (trash_vector); k++)
    trash_vector[k] = (char) rand();

  for (k = 0; k < TRACES; k++)
    {
      for (i = 0; i < BUFLEN; i++)
	vector32[k][i] = rand() % 9;
    }

  samples = TRACES * BUFLEN;

  traces = readtraces();

  /* Burn at least 1s of CPU, to warm up. */
  gettimeofday(&a, NULL);
  for ( ; ; )
    {
      m._sum += trash_cache();
      gettimeofday(&b, NULL);
      if  (b.tv_sec != a.tv_sec &&
	   b.tv_sec != a.tv_sec + 1)
	break;
    }

  m._sum = 0;

  for (init_speed(&m); check_speed(&m, &elapsed_compress16); )
    {
      for (k = 0; k < traces; k++)
	{
	  np16[k] = dptc_compress16(vector16[k], len[k],
				    packed16[k], len[k],
				    DPTC_BITS);
	}
    }

  for (init_speed(&m); check_speed(&m, &elapsed_compress32); )
    {
      for (k = 0; k < traces; k++)
	{
	  np32[k] = dptc_compress32(vector32[k], len[k],
				    packed32[k], len[k],
				    DPTC_BITS);
	}
    }

  for (init_speed(&m); check_speed(&m, &elapsed_ads_compress32); )
    {
      for (k = 0; k < traces; k++)
	{
	  ads_np32[k] = dptc_ads_compress32(ads_vector32[k], len[k],
					    ads_packed32[k], len[k],
					    DPTC_BITS + ADS_MAX_LEVEL,
					    ADS_MAX_LEVEL);
	}
    }

  for (init_speed(&m); check_speed(&m, &elapsed_unpack16); )
    {
      for (k = 0; k < traces; k++)
	{
	  dptc_unpack16(packed16[k], np16[k],
			unpack16[unp_k], len[k],
			DPTC_BITS);
	}
    }

  for (k = (TESTSPEED_UNPACK_SAME_STRUCT ? traces - 1 : 0); k < traces; k++)
    for (i = 0; i < len[k]; i++)
      if (vector16[k][i] != unpack16[unp_k][i])
	fprintf (stderr,
		 "%4d/%4d: %5d != %5d\n",
		 k, i, vector16[k][i], unpack16[unp_k][i]);

  for (init_speed(&m); check_speed(&m, &elapsed_unpack32); )
    {
      for (k = 0; k < traces; k++)
	{
	  dptc_unpack32(packed32[k], np32[k],
			unpack32[unp_k], len[k],
			DPTC_BITS);
	}
    }

  for (k = (TESTSPEED_UNPACK_SAME_STRUCT ? traces - 1 : 0); k < traces; k++)
    for (i = 0; i < len[k]; i++)
      if (vector32[k][i] != unpack32[unp_k][i])
	fprintf (stderr,
		 "%4d/%4d: %5d != %5d\n", k, i,
		 vector32[k][i], unpack32[unp_k][i]);

  for (init_speed(&m); check_speed(&m, &elapsed_ads_unpack32); )
    {
      for (k = 0; k < traces; k++)
	{
	  dptc_ads_unpack32(ads_packed32[k], ads_np32[k],
			    ads_unpack32[unp_k], len[k],
			    DPTC_BITS + ADS_MAX_LEVEL,
			    ADS_MAX_LEVEL);
	}
    }

  for (k = (TESTSPEED_UNPACK_SAME_STRUCT ? traces - 1 : 0); k < traces; k++)
    for (i = 0; i < len[k]; i++)
      if (ads_vector32[k][i] != ads_unpack32[unp_k][i])
	fprintf (stderr,
		 "%4d/%4d: %5d != %5d\n",
		 k, i, ads_vector32[k][i], ads_unpack32[unp_k][i]);

  samples = 0;
  for (k = 0; k < traces; k++)
    samples += len[k];

  printf ("pack16: %.6f s, unpack16: %.6f s, "
	  "pack32: %.6f s, unpack32: %.6f s, "
	  "ads_pack32: %.6f s, ads_unpack32: %.6f s\n",
	  elapsed_compress16, elapsed_unpack16,
	  elapsed_compress32, elapsed_unpack32,
	  elapsed_ads_compress32, elapsed_ads_unpack32);

  printf ("pack16: %.2f ns/sample, unpack16: %.2f ns/sample, "
	  "pack32: %.2f ns/sample, unpack32: %.2f ns/sample, "
	  "ads_pack32: %.2f ns/sample, ads_unpack32: %.2f ns/sample\n",
	  elapsed_compress16 / (samples) / 1.e-9,
	  elapsed_unpack16 / (samples) / 1.e-9,
	  elapsed_compress32 / (samples) / 1.e-9,
	  elapsed_unpack32 / (samples) / 1.e-9,
	  elapsed_ads_compress32 / (samples) / 1.e-9,
	  elapsed_ads_unpack32 / (samples) / 1.e-9);

  printf ("traces: %zd  samples: %zd  sum: %d\n", traces, samples, m._sum);

  return 0;
}
