-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2019, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tb_dptc_config_pkg.all;

entity tb_timing is
  port (clk:      in  std_logic;

        pin_in:   in  std_logic_vector(33 downto 0);

        pin_out:  out std_logic_vector(59 downto 0)
        );
end;

architecture RTL of tb_timing is

  signal ff_in : std_logic_vector(pin_in'range);

  signal ff_out: std_logic_vector(pin_out'range);

  signal u_in  : unsigned(pin_in'range);
  signal u_out : unsigned(pin_out'range);

  signal u_tmp1 : unsigned(15 downto 0);
  signal u_tmp2 : unsigned(31 downto 0);


  signal shift_mul_mask:
      std_logic_vector(38 - 1 downto 0);
  signal shift_mul_mask_reg:
      std_logic_vector(shift_mul_mask'range);

  signal shift_input: unsigned(21 downto 0);
  signal shift_bits : unsigned(5 downto 0);
  
  -- Registered value
  signal shift_input_reg: unsigned(shift_input'range);

  -- Shifted data (fits somewhere in accumulated word).
  signal shifted_data   : std_logic_vector(59 downto 0);
  signal shifted_data_1 : std_logic_vector(59 downto 0);
  signal shifted_data_2 : std_logic_vector(59 downto 0);
  signal shifted_data_3 : std_logic_vector(59 downto 0);
 

begin

    pin_out <= ff_out;

    process(clk)
    begin
        if (rising_edge(clk)) then
            ff_in <= pin_in;
      
            -- ff_out(99 downto 33) <= ff_in(99 downto 33);
        end if;
    end process;

    dptc_gen : if (specialbld = 0) generate
        dptc : entity work.dptc_module
            generic map (predictor => predictor,
                         shift_mul => shift_mul,
                         dc_cd_reg => dc_cd_reg,
                         dc_pd_reg => dc_pd_reg,
                         dc_od_reg => dc_od_reg,
                         ec_mc_reg => ec_mc_reg,
                         ec_cb_reg => ec_cb_reg,
                         op_ss_reg => op_ss_reg)
            port map (    
                ----- INPUTS ------
                clk     => clk,

                reset   => ff_in(17),
                dv_in   => ff_in(16),
                data_in => ff_in(tb_input_width-1 downto 0),
                flush   => ff_in(18),
            
                ----- OUTPUTS -----
                dv_out   => ff_out(32),
                done     => ff_out(33),
                out_word => ff_out(tb_output_width-1 downto 0)
                );
    end generate;

    special_gen : if (specialbld = 1) generate

        u_in   <= unsigned(ff_in);

        -- 16-bit adder
        special_adder : if (specialtype = 0) generate

            u_out(tb_input_width-1 downto 0) <=
                u_in(tb_input_width-1 downto 0) + u_in(16+tb_input_width-1 downto 16);
            
            sa_out: if (tb_input_width < 16) generate
                u_out(15 downto tb_input_width) <= (others => '0');
            end generate;

        end generate;

        -- 16-bit multiplier
        special_multiplier : if (specialtype = 1) generate

            sa_out: if (tb_input_width < 16) generate
                u_tmp1(15 downto tb_input_width) <= (others => '0');
                u_tmp2(31 downto 2*tb_input_width) <= (others => '0');
            end generate;

            u_tmp1(tb_input_width-1 downto 0) <=
                u_in(tb_input_width-1 downto 0); -- + ("000000000000000" & u_in(32));

            u_tmp2(2*tb_input_width-1 downto 0) <=
                u_in(16+tb_input_width-1 downto 16) * u_tmp1(tb_input_width-1 downto 0);

            u_out(15 downto 0) <=
                u_tmp2(tb_input_width-1 downto 0) + ("000000000000000" & u_in(33));
            
        end generate;

        -- 22-bit << 0-37 shifter
        special_shifter : if (specialtype = 2) generate

            u_out(59 downto 0) <=
                ((59 downto 6+tb_input_width => '0') & u_in(6+tb_input_width-1 downto 0)) sll
                to_integer(u_in(22+5 downto 22+0));
            
        end generate;

        -- 22-bit << 0-37 shifter, by multipliers
        special_shift_by_mul : if (specialtype = 3) generate

            shift_input <= (21 downto 6+tb_input_width => '0') &
                           u_in(6+tb_input_width-1 downto 0);
            shift_bits  <= u_in(22+5 downto 22+0);
            
            mul_mask: for i in 0 to shift_mul_mask'left generate
                shift_mul_mask(i) <=
                    '1' when (to_integer(shift_bits) = i) else '0';
            end generate;
            
            process (clk)
            begin
                if rising_edge(clk) then
                    shift_input_reg <= shift_input;
                    shift_mul_mask_reg <= shift_mul_mask;

                    shifted_data_1 <=
                        std_logic_vector(unsigned(shift_input_reg) *
                                         unsigned(shift_mul_mask_reg));
                    shifted_data_2 <= shifted_data_1;
                    shifted_data_3 <= shifted_data_2;
                    shifted_data <= shifted_data_3;
                end if;
            end process;

             u_out(59 downto 0) <= unsigned(shifted_data);

        end generate;

        process(clk)
        begin
            if (rising_edge(clk)) then
                ff_out <= std_logic_vector(u_out);
            end if;
        end process;

    end generate;
    
end;
