-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2019, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;

use work.tb_dptc_config_pkg.all;
use work.all;

entity tb_dptc is
    generic (seed : integer := 0);
end entity tb_dptc;
 
architecture RTL of tb_dptc is

    signal runsim : boolean := true;
  
    signal clk : std_logic;
  
    constant period : time := 10 ns;

    -- Interface with compressor 
    signal reset    : std_logic;
    signal in_data  : std_logic_vector (tb_input_width-1 downto 0);
    signal in_word  : std_logic;
    signal flush    : std_logic;
    
    signal out_data : std_logic_vector (tb_output_width-1 downto 0);
    signal out_word : std_logic;
    signal done     : std_logic;

begin
    dptc: entity work.dptc_module
        generic map (predictor => predictor,
                     shift_mul => shift_mul,
                     dc_cd_reg => dc_cd_reg,
                     dc_pd_reg => dc_pd_reg,
                     dc_od_reg => dc_od_reg,
                     ec_mc_reg => ec_mc_reg,
                     ec_cb_reg => ec_cb_reg,
                     op_ss_reg => op_ss_reg)
        port map (    
            ----- INPUTS ------
            clk      => clk,
            reset    => reset,
            dv_in    => in_word,
            data_in  => in_data,
            flush    => flush,
            
            ----- OUTPUTS -----
            dv_out   => out_word,
            out_word => out_data,
            done     => done
            );

    -- Drive the clock
    process
    begin
        if runsim then
            clk <= '1';
            wait for 5 ns;
            clk <= '0';
            wait for 5 ns;
        else
            wait;
        end if;
    end process;

    -- Read input file with data to compress
    testprocess: process is
        variable row    : line;
        variable rowval : integer;
        variable gotval : boolean;
        variable read_data : std_logic_vector (30-1 downto 0);
        variable l      : line;
        variable cycles_reset_to_data : integer := 0;
        variable cycles_reset_extra : integer := 0;
        variable cycles_end_to_flush : integer := 0;
        variable cycles_done_to_noflush : integer := 0;
        variable cycles_flush_to_reset : integer := 0;
        variable x : std_logic_vector(31 downto 0);
    begin
        write (l, String'("SEED: "));
        write (l, seed);
        writeline (output, l);

        x := std_logic_vector(to_unsigned(seed, x'length));

        reset <= '1';
        in_word <= '0';
        flush <= '0';
        in_data <= std_logic_vector(to_unsigned(0, in_data'length));
        -- Hack: get a bit into the clock cycle, to set the data to hold.
        wait for 0.1 ns;
        -- Wait a few clock cycles to get rid of meta values.
        wait for 60 ns;
        reset <= '0';
    
        wait for 30 ns;
        in_word <= '1';
        --~ wait for period;

        while (not endfile(input)) loop
            readline(input,row);
            read(row,rowval,gotval);
            if (not gotval) then
                in_word <= '0';
                --cycles_end_to_flush := (cycles_end_to_flush + 1) mod 9;
                --wait for cycles_end_to_flush*10 ns;
                flush <= '1';
                while (done = '0') loop
                    wait for 10 ns;
                end loop;
                cycles_done_to_noflush := cycles_done_to_noflush mod 5;
                wait for cycles_done_to_noflush*10 ns;
                flush <= '0';
                cycles_flush_to_reset := cycles_flush_to_reset mod 7;
                wait for cycles_flush_to_reset*10 ns;
                reset <= '1';
                wait for 60 ns;
                cycles_reset_extra := cycles_reset_extra mod 7;
                wait for cycles_reset_extra*10 ns;
                reset <= '0';
                cycles_reset_to_data := cycles_reset_to_data mod 7;
                wait for cycles_reset_to_data*10 ns;
                in_word <= '1';
                write (l, String'("OUTEND"));
                writeline (output, l);
            else
                cycles_reset_to_data  := cycles_reset_to_data+(rowval mod 17);
                cycles_reset_extra    := cycles_reset_extra+(rowval mod 13);
                cycles_end_to_flush   := cycles_end_to_flush+(rowval mod 11);
                cycles_done_to_noflush:= cycles_done_to_noflush+(rowval mod 5);
                cycles_flush_to_reset := cycles_flush_to_reset+(rowval mod 7);
                read_data := std_logic_vector(to_unsigned(rowval, 30));
                in_data <= read_data(in_data'range);
                --write (l, String'("INPUT: 0x"));
                --hwrite (l, read_data);
                --write (l, String'(" = "));
                --write (l, read_data);
                --writeline (output, l);
                wait for 10 ns;

                -- A seed of 0 will make the xorshift not change at all,
                -- and stay 0.
                
                x := x xor (  x(31-13 downto 0) & "0000000000000");
                x := x xor ("00000000000000000" & x(31 downto 17));
                x := x xor (   x(31-5 downto 0) & "00000"        );
                
                while (x(0) = '1') loop
                    x := x xor (  x(31-13 downto 0) & "0000000000000");
                    x := x xor ("00000000000000000" & x(31 downto 17));
                    x := x xor (   x(31-5 downto 0) & "00000"        );

                    -- Do a cycle with no data.
                    in_word <= '0';
                    wait for 10 ns;
                end loop;
                in_word <= '1'; -- in case it was set to 0

            end if;
        end loop;

        -- We are done providing data values
        in_word <= '0';
        flush <= '1';
        -- Due to pipelining of compressor, we have to wait for
        -- two cycles until we request the last bits
        wait for 100 ns;
        flush <= '0';
        wait for 100 ns;
        write (l, String'("OUTEND"));
        writeline (output, l);

        -- Terminate simulation.
        runsim <= false;
        wait;
    
        assert (FALSE) report "Simulation end." severity failure;

    end process testprocess;

    -- Write produced data words
    process
        variable l : line;
    begin       
        if runsim then
            -- Hack: get into the clock cycle, to get the latched data.
            wait for 1 ns;
            if (out_word = '1') then
                write (l, String'("OUTPUT: 0x"));
                hwrite (l, out_data);
                write (l, String'(" = "));
                write (l, out_data);
                writeline (output, l);
            end if;
            wait for 9 ns;
        else
            wait;
        end if;
    end process;

end architecture RTL;
