-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2017, Giovanni Bruni
-- Copyright (c) 2019, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.dptc_util_pkg.all;

entity dptc_diff_calculator is
    generic (predictor : integer := 0;
             cd_reg : integer := 0;
             pd_reg : integer := 0;
             od_reg : integer := 0
             );
    port (clk:      in  std_logic;
          reset:    in  std_logic;
          -- Input value.
          data_in:  in  std_logic_vector;
          -- Input value is valid.
          dv_in:    in  std_logic;
          -- Flush pipeline.
          flush:    in  std_logic;

          -- Output value.
          out_diff: out std_logic_vector;
          -- Output data is valid.
          out_dv:   out std_logic;
          -- Flush pipeline.
          out_flush: out std_logic
    );
end;

architecture RTL of dptc_diff_calculator is
    
    -- Current value.
    signal cur_data: unsigned(data_in'range);
    -- Previous value.
    signal prev_data: unsigned(data_in'range) := (others => '0');

    -- Difference of current and previous value.
    signal cur_diff: signed(data_in'range);
    -- Pipeline register
    signal cur_diff_reg1: signed(data_in'range) := (others => '0');
    -- Associated pipeline
    signal dv_reg1: std_logic;
    signal flush_reg1: std_logic;

    -- Negative of current difference value.  (Used to find if positive.)
    signal minus_cur_diff: signed(data_in'range) := (others => '0');

    -- Current difference is negative or positive?  (0 is neither)
    signal cur_neg, cur_pos: std_logic := '0';

    -- Shift register for the last three differences being negative
    -- (or positive).
    signal prev_neg, prev_pos: std_logic_vector(2 downto 0) := (others => '0');
    -- Last three differences were negative (or positive).
    signal last3_neg, last3_pos: std_logic := '0';
    -- Last three differences had same sign.
    signal last3_same_sign: std_logic := '0';
    
    -- Previous difference.
    signal prev_diff: signed(data_in'range) := (others => '0');
    -- Difference after applying the predictor.
    signal pred_diff: signed(data_in'range);
    -- Pipeline register
    signal pred_diff_reg2: signed(data_in'range) := (others => '0');
    -- Associated pipeline
    signal dv_reg2: std_logic;
    signal flush_reg2: std_logic;

    -- Should next difference be flipped?
    signal flip: std_logic := '0';
    -- Flipped difference value (depending on sign of previous difference).
    signal diff_flip: signed(data_in'range);
    -- Alias
    signal diff_flip_v: std_logic_vector(data_in'range);
    -- Pipeline register
    signal diff_flip_v_reg3: std_logic_vector(data_in'range) :=
        (others => '0');
    -- Associated pipeline
    signal dv_reg3: std_logic;
    signal flush_reg3: std_logic;
    
begin

    cur_data <= unsigned(data_in);

    cur_diff <= signed(cur_data - prev_data);

    dptc_ppl(p=>cd_reg,c=>clk,i=>cur_diff,   o=>cur_diff_reg1);
    dptc_ppl(p=>cd_reg,c=>clk,i=>dv_in,      o=>dv_reg1);
    dptc_ppl(p=>cd_reg,c=>clk,i=>flush,      o=>flush_reg1);

    use_pred: if (predictor = 1) generate
        -- Current difference is negative if sign bit is set.
        cur_neg <= cur_diff_reg1(cur_diff'left);
        -- Check for positive by
        -- checking the sign bit of -(current difference).
        -- It is not really correct, as the most negative value will also
        -- report as positive, but this then matches the operation in the
        -- C code unpacker.
        minus_cur_diff <= -cur_diff_reg1;
        cur_pos <= minus_cur_diff(minus_cur_diff'left);

        last3_neg <= prev_neg(2) and prev_neg(1) and prev_neg(0);
        last3_pos <= prev_pos(2) and prev_pos(1) and prev_pos(0);

        last3_same_sign <= last3_neg or last3_pos;

        pred_diff <=
            cur_diff_reg1 - prev_diff
            when (last3_same_sign = '1') else cur_diff_reg1;
    end generate;
    no_pred: if (predictor = 0) generate
        pred_diff <= cur_diff_reg1;
    end generate;

    -- Inserting a pipeline stage after pred_diff helps timing.
    -- Better than after diff_calculator, as it combines with first stage
    -- in the encoder.
    dptc_ppl(p=>pd_reg,c=>clk,i=>pred_diff,  o=>pred_diff_reg2);
    dptc_ppl(p=>pd_reg,c=>clk,i=>dv_reg1,    o=>dv_reg2);
    dptc_ppl(p=>pd_reg,c=>clk,i=>flush_reg1, o=>flush_reg2);
    
    diff_flip <= pred_diff_reg2 when (flip = '0') else -pred_diff_reg2;

    diff_flip_v <= std_logic_vector(diff_flip);

    process(clk)
    begin
        if rising_edge(clk) then
            if (dv_in = '1') then
                prev_data <= cur_data;
            end if;

            if (dv_reg1 = '1') then
                prev_neg(prev_neg'range) <=
                    prev_neg(prev_neg'left-1 downto 0) & cur_neg;
                prev_pos(prev_pos'range) <=
                    prev_pos(prev_pos'left-1 downto 0) & cur_pos;

                prev_diff <= cur_diff_reg1;
            end if;

            if (dv_reg2 = '1' and
                pred_diff_reg2 /= 0) then
                flip <= pred_diff_reg2(pred_diff_reg2'left);
            end if;

            -- Reset overrides assignments above.
            if (reset = '1') then
                prev_neg <= (others => '0');
                prev_pos <= (others => '0');

                prev_diff <= (others => '0');
                prev_data <= (others => '0');
                flip <= '0';
            end if;
            
        end if;
    end process;

    -- Inserting a pipeline stage after the diff_calculator helps timing.
    dptc_ppl(p=>od_reg,c=>clk,i=>dv_reg2,    o=>dv_reg3);
    dptc_ppl(p=>od_reg,c=>clk,i=>flush_reg2, o=>flush_reg3);
    dptc_ppl(p=>od_reg,c=>clk,i=>diff_flip_v,o=>diff_flip_v_reg3);

    out_dv    <= dv_reg3;
    out_flush <= flush_reg3;
    out_diff  <= diff_flip_v_reg3;
    
end RTL;
