-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2017, Giovanni Bruni
-- Copyright (c) 2019, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.dptc_util_pkg.all;

entity dptc_encoding is
    generic (chunk_size: natural;
             mc_reg: integer := 0;
             cb_reg: integer := 0
             );
    port (clk:      in  std_logic;
          reset:    in  std_logic;
          -- Input value.
          data_in:  in  std_logic_vector;
          -- Input value is valid.
          dv_in:    in  std_logic;
          -- Flush pipeline.
          flush:    in  std_logic;

          -- Masked value to store.          
          out_encoded:   out std_logic_vector;
          -- Number of bits (of current chunk) to store.
          out_num_bits:  out std_logic_vector;
          -- Signals the presence of data to store.
          out_dv:        out std_logic;
          -- Start new chunk together with this value.
          out_new_chunk: out std_logic;
          -- Flush pipeline.
          out_flush: out std_logic;
          -- Step bits pipeline.
          out_step_info: out std_logic_vector(1 downto 0)
       );
end;

architecture RTL of dptc_encoding is

    -- Number of bits needed to represent the number of used data bits.
    constant width_bits: natural := out_num_bits'length;
    
    -- Width of counter for new chunk recognition (chunk_size power of 2).
    constant cnk_cnt_width: natural := dptc_log2(chunk_size);

    -- Counter for new chunk, active mask, etc.  On input side of pipeline.
    signal chunk_count_in: unsigned(cnk_cnt_width-1 downto 0) :=
        (others => '0');
    -- Counter for new chunk, active mask, etc.  On output side of pipeline.
    signal chunk_count_out: unsigned(cnk_cnt_width-1 downto 0) :=
        (others => '0');
    
    type chunk_pipeline is
        array (chunk_size downto 0) of std_logic_vector(data_in'range);
    -- Data pipeline.
    signal c_pipe: chunk_pipeline := (others => (data_in'range => '0'));

    -- Data valid pipeline.
    signal dv_pipe: std_logic_vector(chunk_size downto 0) := (others => '0');
    
    -- Sign bit of input data.
    signal data_in_sign: std_logic;

    -- Mask for current input value.
    signal mask_cur_ext: std_logic_vector(data_in'left + 1 downto
                                          data_in'right);
    -- Mask for current input value.
    signal mask_cur: std_logic_vector(data_in'range);
    -- Pipeline register
    signal mask_cur_reg: std_logic_vector(data_in'range) := (others => '0');
    -- Associated pipelines of related values.
    signal data_in_reg: std_logic_vector(data_in'range) := (others => '0');
    signal dv_in_reg: std_logic;
    signal flush_reg: std_logic;
    
    -- Accumulated mask, including new input value mask.
    signal mask_or_cur: std_logic_vector(data_in'range);
    -- Mask to keep next clock cycle.  Accumulated, fresh or previous.
    signal mask_next: std_logic_vector(data_in'range);

    -- Accumulated mask.
    signal mask: std_logic_vector(data_in'range) := (others => '0');

    -- Latched mask for values that are leaving pipeline.
    signal active_mask: std_logic_vector(data_in'range) := (others => '0');
    -- To use as next latched mask.
    signal active_mask_next: std_logic_vector(data_in'range);
    -- Most negative value representable within mask, used as bias.
    signal active_mask_base: std_logic_vector(data_in'range);

    -- Alias of active mask used for counting (lowest fixed '1' bit ignored)
    signal active_mask_to_count: std_logic_vector(data_in'left downto 1);
    -- Number of bits in active mask.
    signal active_mask_num_bits: std_logic_vector(out_num_bits'range);

    -- Alias of next active mask used for counting
    signal active_mask_next_to_count: std_logic_vector(data_in'left downto 1);
    -- Number of bits in next active mask.
    signal active_mask_next_num_bits: std_logic_vector(out_num_bits'range);
    -- Latched value of number of bits in next active mask.
    signal active_mask_num_bits_reg: std_logic_vector(out_num_bits'range) :=
        (others => '0');

    -- Value from pipeline.
    signal out_temp: std_logic_vector(data_in'range);
    -- Value from pipeline is valid.
    signal dv_temp: std_logic;

    -- Value pushed out from pipeline
    signal dv_pipe_out: std_logic;

    -- There is some valid data in the pipeline.
    signal any_dv: std_logic;

    -- Start new chunk together with this value.
    signal new_chunk: std_logic := '0';
    
    -- Value from pipeline, biased with most negative value.
    signal encoded_biased: std_logic_vector(data_in'range);

begin

    data_in_sign <= data_in(data_in'left);

    -- Generate the mask based on the current data.
    mask_cur_ext(mask_cur_ext'left) <= '0';
    mask_cur_gen: for i in data_in'range generate
        mask_cur_ext(i) <= mask_cur_ext(i+1) or (data_in_sign xor data_in(i));
    end generate;

    mask_cur <= mask_cur_ext(mask_cur'range);

    dptc_ppl(p=>mc_reg,c=>clk,i=>mask_cur,       o=>mask_cur_reg);
    dptc_ppl(p=>mc_reg,c=>clk,i=>data_in,        o=>data_in_reg);
    dptc_ppl(p=>mc_reg,c=>clk,i=>dv_in,          o=>dv_in_reg);
    dptc_ppl(p=>mc_reg,c=>clk,i=>flush,          o=>flush_reg);
    
    mask_or_cur <= mask or mask_cur_reg;
    
    mask_next <=
        mask         when (dv_in_reg = '0') else
        mask_cur_reg when (chunk_count_in = 1) else
        mask_or_cur;

    active_mask_next <=
        mask         when (dv_in_reg = '0') else
        mask_or_cur;

    -- Highest set bit (most negative value) of active mask.
    active_mask_base <=
        active_mask and not ('0' & active_mask(active_mask'left downto 1));

    encoded_biased <=
        std_logic_vector(unsigned(out_temp) - unsigned(active_mask_base));
    out_encoded <= (encoded_biased and active_mask);
    
    active_mask_to_count <=
        active_mask(active_mask'left downto 1);

    count_bits: entity work.dptc_count_low_bits
        port map (x => active_mask_to_count,
                  y => active_mask_num_bits);

    -- The following are used as an alternative to the calculation above.
    -- The gain is (sometimes, but not always) to having the count value
    -- latched.  Less logics in the output packing routing in the same
    -- cycle.  Not always a win, therefore optional.
    active_mask_next_to_count <=
        active_mask_next(active_mask_next'left-1 downto 0);

    count_bits_next: entity work.dptc_count_low_bits
        port map (x => active_mask_next_to_count,
                  y => active_mask_next_num_bits);

    -- Select whether to use the direct or the latched value.
    out_num_bits <=
        active_mask_num_bits when (cb_reg = 0) else
        active_mask_num_bits_reg;

    c_pipe(c_pipe'left) <= data_in_reg;
    dv_pipe(dv_pipe'left) <= dv_in_reg;

    out_temp <= c_pipe(0);
    dv_temp <= dv_pipe(0);

    dv_pipe_out <= dv_temp and (dv_in_reg or flush_reg);

    out_dv <= dv_pipe_out;

    any_dv <= or_reduction(dv_pipe);

    out_flush <= flush_reg and not any_dv;

    -- chunk_count_next <= ('0' & chunk_count) + dv_pipe_out;

    process (clk)        
    begin
        if rising_edge(clk) then
            if (dv_in_reg = '1' or flush_reg = '1') then
                c_pipe(c_pipe'left-1 downto 0) <=
                    c_pipe(c_pipe'left downto 1);

                dv_pipe(dv_pipe'left-1 downto 0) <=
                    dv_pipe(dv_pipe'left downto 1);
            end if;

            if (reset = '1') then
                dv_pipe(dv_pipe'left-1 downto 0) <= (others => '0');
            end if;

            mask <= mask_next;
            
            if (chunk_count_out = 0 and dv_pipe_out = '1') then
                -- Effectively shift mask up by one bit, gives twice the
                -- range, to also store sign.
                active_mask <= active_mask_next(mask'left-1 downto 0) & '1';
                active_mask_num_bits_reg <= active_mask_next_num_bits;
                new_chunk <= '1';
            elsif (dv_pipe_out = '1') then
                new_chunk <= '0';
            end if;

            chunk_count_in  <= chunk_count_in  + ("" & (dv_in_reg or flush_reg));

            chunk_count_out <= chunk_count_out + ("" & dv_pipe_out);

            if (reset = '1') then
                -- chunk_count <= (0 => '0', others => '0');
                chunk_count_in  <= (others => '0');
                chunk_count_out <= (others => '0');
                new_chunk <= '0';
                -- Prepare active mask for first item, which is stored
                -- fully.
                active_mask <= (others => '1');
                active_mask_num_bits_reg <=
                    std_logic_vector(to_unsigned(data_in'length-1,
                                                 out_num_bits'length));
            end if;
        end if;
    end process;

    out_new_chunk <= new_chunk;

end RTL;
