-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2017, Giovanni Bruni
-- Copyright (c) 2019, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.dptc_util_pkg.all;

entity dptc_module is
    generic (predictor : integer := 0;
             shift_mul : integer := 0; -- Must be 0 - see comment in
                                       -- dptc_shift_accumulate.
             dc_cd_reg : integer := 0;
             dc_pd_reg : integer := 0;
             dc_od_reg : integer := 0;
             ec_mc_reg : integer := 0;
             ec_cb_reg : integer := 0;
             op_ss_reg : integer := 0
             );
    port (clk:      in  std_logic;
          reset:    in  std_logic;
          -- Input value.
          data_in:  in  std_logic_vector;
          -- Input value is valid.
          dv_in:    in  std_logic;
          -- Flush pipeline.
          flush:    in  std_logic;

          -- Signals the presence of a valid output word.
          dv_out:   out std_logic;
          -- Output data.
          out_word: out std_logic_vector;
          -- No further data words will be emitted after this cycle.
          done:     out std_logic
    );
end;


architecture RTL of dptc_module is

    -- The number of values in each chunck is fixed in the data format.
    constant chunk_size: natural := 4;

    -- Number of bits needed to represent the number of used data bits.
    constant width_bits: natural := dptc_log2(data_in'length-1) + 1;

    -- Alias for input data, with 0 as LSB index.
    signal data_in_align: std_logic_vector(data_in'length-1 downto 0);
    -- Alias for output data, with 0 as LSB index.
    signal out_word_align: std_logic_vector(out_word'length-1 downto 0);

    -- Difference value to encode from difference/predictor unit.
    signal out_diff: std_logic_vector(data_in_align'range);
    -- Value from differnce/predictor is valid (pipelined from input).
    signal dv_diff: std_logic;
    -- Flush pipeline (pipelined).
    signal flush_diff: std_logic;

    -- Masked value from chunk encoder to store.
    signal encoded: std_logic_vector(data_in_align'range);
    -- Number of bits (of current chunk) to store.
    signal num_bits: std_logic_vector(width_bits-1 downto 0);
    -- Value valid (should be stored).
    signal dv_enc: std_logic;
    -- Start new chunk (store number of bits) together with this value.
    signal new_chunk: std_logic;
    -- Flush pipeline (pipelined).
    signal flush_enc: std_logic;

begin

    data_in_align <= data_in; 
    
    diff_calc: entity work.dptc_diff_calculator
        generic map(predictor => predictor,
                    cd_reg => dc_cd_reg,
                    pd_reg => dc_pd_reg,
                    od_reg => dc_od_reg)
        port map (    
            -- Inputs:
            clk       => clk,
            reset     => reset,
            data_in   => data_in_align,
            dv_in     => dv_in,
            flush     => flush,
            -- Outputs:
            out_diff  => out_diff,
            out_dv    => dv_diff,
            out_flush => flush_diff
            );
    
    
    encoding: entity work.dptc_encoding
        generic map (chunk_size => chunk_size,
                     mc_reg => ec_mc_reg,
                     cb_reg => ec_cb_reg)
        port map (    
            -- Inputs:
            clk       => clk,
            reset     => reset,
            data_in   => out_diff,
            dv_in     => dv_diff,
            flush     => flush_diff,
            -- Outputs:
            out_num_bits  => num_bits,
            out_new_chunk => new_chunk,
            out_encoded   => encoded,
            out_dv        => dv_enc,
            out_flush     => flush_enc
            );
                
    output_packing: entity work.dptc_output_packing
        generic map (shift_mul => shift_mul,
                     ss_reg => op_ss_reg)
        port map (    
            -- Inputs:
            clk       => clk,
            reset     => reset,
            in_bits   => num_bits,
            new_chunk => new_chunk,
            data_in   => encoded,
            dv_in     => dv_enc,
            flush     => flush_enc,
            -- Outputs:
            out_word  => out_word_align,
            out_dv    => dv_out,
            done      => done
            );

    out_word <= out_word_align;
    
end RTL;
