-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2017, Giovanni Bruni
-- Copyright (c) 2019, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.dptc_util_pkg.all;

entity dptc_output_packing is
    generic (shift_mul : integer := 0;
             ss_reg : integer := 0
             );
    port (clk:       in  std_logic;
          reset:     in  std_logic;
          -- Word describing how many bits are needed for each data.
          in_bits:   in  std_logic_vector;
          -- Signal for the beginning of a new chunk.
          new_chunk: in  std_logic;
          -- Data to be stored.
          data_in:   in  std_logic_vector;
          -- Data valid signal.
          dv_in:     in  std_logic;
          -- Flush pipeline.
          flush:     in  std_logic;

          -- Output data.
          out_word:  out std_logic_vector;
          -- Signals the presence of a valid output word.
          out_dv:    out std_logic;
          -- No further data words will be emitted after this cycle.
          done:      out std_logic
    );
end;

-- The maximum number of bits that need to be handled as overflow bits
-- is given by the longest possible output.
--
-- The longest possible output is generated when we emit a long header.
-- This will be followed by (chunk_size-1) entries that have no header,
-- they thus act to reduce the amount of overflow (gradually).
--
-- If the *average* of the chunk item with a chunk header and the
-- remaining chunk items can exceed 32, an additional issue would be
-- that multiple output words would need to be generated per cycle.
-- This is not supported.
--
-- The average number of bits emitted is:
--
-- max_avg_bits = step_bits + data_bits + (2 + long_bits) / chunk_size
--
-- With step_bits = 2, chunk_size = 4, and data_bits = 28,
-- yields max_avg_bits = 31.75 <= 32.
--
-- The maximum overflow would occur during a sequence which started
-- with 31 overflow bits from the previous value:
--
-- chunk header: 31 + (2+28+2+5) => 68 ; 32 are emitted, 36 are overflow
-- chunk value:  36 + (2+28)     => 66 ; 32 are emitted, 34 are overflow
-- chunk value:  34 + (2+28)     => 64 ; 32 are emitted, 32 are overflow
-- chunk value:  32 + (2+28)     => 62 ; 32 are emitted, 30 are overflow
-- chunk header: 30 + (2+28+2+5) => 67 ; better than above, so ok
--
-- The maximum overflow is thus one bit less than the maximum generated
-- word.
--
-- Also to take into consideration would be the first word.  But if this
-- is below 32+31=63 bits, then it cannot cause more overflow to be
-- carried into the next word than can already be handled worst case
-- above.

architecture RTL of dptc_output_packing is

    -- Number of bits needed to store the full data value.
    constant data_bits: natural := data_in'length;

    -- Number of bits needed to represent the number of used data bits.
    constant width_bits: natural := in_bits'length;
    -- Number of bits-1 needed to represent the number of used output bits.
    constant output_log2: natural := dptc_log2(out_word'length);

    -- Number of bits needed to store for long encoding of bits difference.
    constant long_bits: natural := dptc_log2(data_bits-3-1)+1;

    -- Maximum number of bits for chunk header.
    constant max_chunk_header_sz: natural := 2 + long_bits;
    -- Maximum number of bits stored by one chunk.
    constant max_store_chunk_sz: natural :=
        max_chunk_header_sz + data_bits;
    -- Maximum number of bits combined with accumulation.
    constant max_accum_sz: natural :=
        out_word'length-1 + max_store_chunk_sz;
    -- Maximum shift (plus 1)
    constant max_shift_val: natural :=
        out_word'length + max_chunk_header_sz;

    -- Number of bits needed to represent the number of used output bits.
    constant max_accum_sz_log2: natural := dptc_log2(max_accum_sz);

    -- Output word is valid;
    signal dv_out: std_logic := '0';

    -- Accumulation overflowed, output word is valid.
    signal has_overflow: std_logic;
    -- Accumulation overflowed last cycle.
    signal has_overflow_prev: std_logic := '0';

    signal dv_out_temp1: std_logic;
    signal dv_out_temp: std_logic;
    
    -- There are remaining bits in accumulated output.
    signal has_partial: std_logic;
    -- Last cycle has remaining bits in accumulated output.
    signal has_partial_prev: std_logic := '0';

    -- Pipelined flush request.
    signal flush_prev: std_logic := '0';
    -- Output any last data word.
    signal do_flush: std_logic;

    -- We have flushed, i.e. emitted any last data word.
    signal flushed: std_logic := '0';
    -- Pipelined.
    signal flushed_prev: std_logic := '0';

    -- Number of bits for chunk encoded last time.
    signal prev_bits: std_logic_vector(width_bits-1 downto 0) :=
        (others => '0');
    -- Change in number of bits to encode for chunk.
    -- signal diff_bits: unsigned(width_bits-1 downto 0);

    -- Can store values -2*n .. 2*n-1
    signal ext_diff_short_bits: signed(width_bits+2-1 downto 0);
    signal unwrap_diff_short_bits: signed(width_bits+2-1 downto 0);
    signal diff_short_bits: unsigned(width_bits-1 downto 0);

    signal ext_diff_long_bits: signed(width_bits+2-1 downto 0);
    signal unwrap_diff_long_bits: signed(width_bits+2-1 downto 0);
    signal diff_long_bits: unsigned(long_bits-1 downto 0);

    -- Store number of bits as short (2-bit) encoding (diff -1,0 or 1).
    signal store_short: std_logic;

    -- Bits to store for chunk header.
    signal header_add_masked: unsigned(max_accum_sz_log2 downto 0);
    -- Additional bits to store for long encoding of number of bits in chunk.
    signal in_bits_add_masked: unsigned(max_accum_sz_log2 downto 0);

    -- Pipeline register
    signal store_short_reg: std_logic;
    signal header_add_masked_reg: unsigned(max_accum_sz_log2 downto 0);
    signal in_bits_add_masked_reg: unsigned(max_accum_sz_log2 downto 0);
    -- Associated pipeline
    signal do_flush_reg: std_logic;
    signal in_bits_reg: std_logic_vector(in_bits'range);
    signal dv_in_reg: std_logic;
    signal data_in_reg: std_logic_vector(data_in'range);
    signal diff_short_bits_reg: unsigned(diff_short_bits'range);
    signal diff_long_bits_reg: unsigned(diff_long_bits'range);
    signal new_chunk_reg: std_logic;

    -- Numbers of bits accumulated, from last round (excluding overflow).
    signal bits_used_add_src: unsigned(max_accum_sz_log2 downto 0);
    -- Number of bits accumulated, together with new value.
    signal bits_used_next: unsigned(max_accum_sz_log2 downto 0);

    -- Numbers of bits accumulated, high bit marks overflow (= output).
    signal bits_used: unsigned(max_accum_sz_log2 downto 0) := (others => '0');

    -- Data for chunk header (either using short or long encoding).
    signal header_bits: std_logic_vector(max_chunk_header_sz - 1 downto 0);

    -- Actual input for shifter.
    signal shift_input: std_logic_vector(max_store_chunk_sz - 1 downto 0);

    -- Number of bits to shift current value (and chunk header).
    signal shift_bits: unsigned(max_accum_sz_log2 downto 0);

    -------------------------------------------------------------------

begin

    out_dv <= dv_out;

    done <= flushed_prev;

    do_flush <= (not dv_in) and flush_prev;

    dptc_ppl(p=>ss_reg,c=>clk,i=>do_flush,       o=>do_flush_reg);

    -- Delay the dv signal by one clock cycle.
    -- The data is also delayed (at least) by one.
    process (clk)
    begin
        if rising_edge(clk) then
            if (do_flush_reg = '1') then
                flushed <= '1';
            end if;

            if reset = '1' then
                flushed <= '0';
            end if;

            has_overflow_prev <= has_overflow;
            has_partial_prev <= has_partial;

            dv_out <= dv_out_temp;
            flush_prev <= flush;
            flushed_prev <= flushed;
        end if;
    end process;

    has_overflow <= bits_used(bits_used'left);

    has_partial <=
        '1' when (to_integer(bits_used(bits_used'left - 1 downto 0)) /= 0)
        else '0';

    -- Output a data word.  Either if overfull.  Or if we shall flush
    -- and had remaining bits.  The latter is one cycle later, in case
    -- last encoded value causes both.
    dv_out_temp1 <=
        has_partial_prev when (do_flush_reg = '1') else has_overflow;
    dv_out_temp <=
        dv_out_temp1 and (not flushed) and (not reset);

    -- Calculate difference with bias of 2 (for easier store and compare).
    -- diff_bits <= (unsigned(in_bits) - unsigned(prev_bits)) + 2;

    ext_diff_short_bits <=
        (signed("00" & in_bits) - signed("00" & prev_bits)) + 2;
    -- Minimum value: 1-n+2 = -n+3  Maximum value: n-1+2=n+1
    -- Minimum value: 0-(n-1)+2 = -n+3  Maximum value: (n-1)+2=n+1

    unwrap_diff_short_bits <=
        (ext_diff_short_bits - data_bits) when (ext_diff_short_bits >=
                                                data_bits) else
        (ext_diff_short_bits + data_bits) when (ext_diff_short_bits <
                                                0) else
        ext_diff_short_bits;

    diff_short_bits <= unsigned(unwrap_diff_short_bits(diff_short_bits'range));

    ext_diff_long_bits <=
        (signed("00" & in_bits) - signed("00" & prev_bits)) - 2;
    -- Minimum value: 1-n-2 = -n-1.  When unwrapped: -1, so still not
    -- withing range.  However, -1 does not correspond to a short encoding
    -- so also needs to be handled.

    unwrap_diff_long_bits <=
        (ext_diff_long_bits -   data_bits) when (ext_diff_long_bits >=
                                                 data_bits) else
        (ext_diff_long_bits + 2*data_bits) when (ext_diff_long_bits <
                                                 -data_bits) else
        (ext_diff_long_bits +   data_bits) when (ext_diff_long_bits <
                                                 0) else
        ext_diff_long_bits;

    diff_long_bits <= unsigned(unwrap_diff_long_bits(diff_long_bits'range));

    -- Use short encoding if diff is -1,0, or 1, i.e. 1,2, or 3 after +2.
    store_short <=
        '1' when (diff_short_bits(diff_short_bits'left downto 2) = 0 and
                  diff_short_bits(1 downto 0) /= 0) else '0';

    header_add_masked <=
        to_unsigned(2,header_add_masked'length) when
        (new_chunk = '1') else (others => '0');
    in_bits_add_masked <=
        to_unsigned(long_bits,in_bits_add_masked'length) when
        (new_chunk = '1' and store_short = '0') else (others => '0');

    dptc_ppl(p=>ss_reg,c=>clk,i=>store_short,    o=>store_short_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>new_chunk,      o=>new_chunk_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>in_bits,        o=>in_bits_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>dv_in,          o=>dv_in_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>data_in,        o=>data_in_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>diff_short_bits,o=>diff_short_bits_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>diff_long_bits, o=>diff_long_bits_reg);

    dptc_ppl(p=>ss_reg,c=>clk,i=>header_add_masked,o=>header_add_masked_reg);
    dptc_ppl(p=>ss_reg,c=>clk,i=>in_bits_add_masked,o=>in_bits_add_masked_reg);

    bits_used_add_src <= '0' & bits_used(bits_used'left - 1 downto 0);

    -- Calculate new position where to put the next word
    bits_used_next <=
        (bits_used_add_src + unsigned(in_bits_reg)) +
        (header_add_masked_reg + in_bits_add_masked_reg + 1);

    header_bits <=
        (std_logic_vector(diff_short_bits_reg(1 downto 0)) &
         (diff_long_bits_reg'range => '0')) when (store_short_reg = '1') else
        (std_logic_vector(diff_long_bits_reg) & "00");

    shift_input <=
        (others => '0') when (reset = '1' or dv_in_reg = '0') else
        (data_in_reg & header_bits) when (new_chunk_reg = '1') else
        (data_in_reg & (header_bits'range => '0'));

    shift_bits <= bits_used_add_src +
                  header_add_masked_reg + in_bits_add_masked_reg;

    process (clk)
    begin
        if rising_edge(clk) then

            if (dv_in = '1') then
                prev_bits <= in_bits;
            end if;

            if (dv_in_reg = '1') then
                -- Storing new position.
                bits_used <= bits_used_next;
            else
                bits_used(bits_used'left) <= '0';
            end if;

            if (reset = '1') then
                bits_used <= (others => '0');
                prev_bits <= (others => '0');
            end if;
        end if;
    end process;

    shift_accumulate: entity work.dptc_shift_accumulate
        generic map(shift_mul       => shift_mul,
                    max_accum_sz    => max_accum_sz,
                    ignore_low_bits => max_chunk_header_sz,
                    max_shift       => max_shift_val)
        port map(clk         => clk,
                 reset       => reset,
                 shift_input => shift_input,
                 shift_bits  => shift_bits,
                 new_word    => has_overflow_prev,
                 out_word    => out_word);
    
end RTL;
