-- -*- vhdl-basic-offset: 4 -*-

-- Copyright (c) 2017, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.dptc_util_pkg.all;

entity dptc_shift_accumulate is
    generic (shift_mul : integer; -- Do not use - see comment in code below!
             max_accum_sz: natural;
             -- Number of low bits of shifted result to discard.
             ignore_low_bits: natural;
             max_shift: natural);
    port (clk:       in  std_logic;
          reset:     in  std_logic;

          -- Input word for shifter, zero-extended.
          shift_input: in std_logic_vector;
          -- Number of bits to shift current value (and chunk header).
          shift_bits:in  unsigned;

          -- Accumulation produces an output word, so keep high part.
          new_word:  in  std_logic;

          -- Output data.
          out_word:  out std_logic_vector
    );
end;

architecture RTL of dptc_shift_accumulate is

    -- Accumulated word is either the additional bits over a full
    -- (emitted) output word, or the output word minus 1.
    constant acc_bits: natural := dptc_max(max_accum_sz - out_word'length,
                                           out_word'length - 1);

    -- Input word for shifter, zero-extended.
    signal shift_input_zext:
        std_logic_vector(max_accum_sz + ignore_low_bits - 1 downto 0);

    signal shift_mul_mask:
        std_logic_vector(max_shift - 1 downto 0) := (others => '0');
    signal shift_mul_mask_reg:
        std_logic_vector(shift_mul_mask'range) := (others => '0');
        
    -- Registered value
    signal shift_input_reg: std_logic_vector(shift_input'range) :=
        (others => '0');
    -- Registered value
    signal shift_input_zext_reg: std_logic_vector(shift_input_zext'range) :=
        (others => '0');
    -- Registered value
    signal shift_bits_reg: unsigned(shift_bits'range) := (others => '0');

    -- Shifted data (fits somewhere in accumulated word).
    signal shifted_data:
        std_logic_vector(shift_input_zext'range) := (others => '0');
    signal shifted_data_1:
        std_logic_vector(shift_input_zext'range) := (others => '0');
    signal shifted_data_2:
        std_logic_vector(shift_input_zext'range) := (others => '0');
    signal shifted_data_3:
        std_logic_vector(shift_input_zext'range) := (others => '0');

    -- Accumulated word.
    signal acc_word: std_logic_vector(acc_bits - 1 downto 0) :=
        (others => '0');

    -- Previously accumulated word, zero-extended.
    signal acc_word_ext:
        std_logic_vector(max_accum_sz - 1 downto 0);
    -- Accumulated word together with new value.
    signal acc_out:
        std_logic_vector(max_accum_sz - 1 downto 0);

begin

    shift_input_zext <=
        (shift_input_zext'left downto shift_input'left+1 => '0') &
        shift_input;

    use_shift: if (shift_mul = 0) generate
        process (clk)
        begin
            if rising_edge(clk) then
                shift_input_zext_reg <= shift_input_zext;
                shift_bits_reg <= shift_bits;

                -- Shifting.  (Expensive part!)
                shifted_data <=
                    std_logic_vector(unsigned(shift_input_zext_reg) sll
                                     to_integer(shift_bits_reg));
            end if;
        end process;
    end generate;

    -- Note: the shift-by-multipler option is currentely not usable!
    --
    -- It introcudes several pipeline stages, which are not accounted for
    -- in the other code (dptc_output_packing).  They were introduced to
    -- allow the toolchains to efficiently use multiplier units, as they
    -- often require built-in register stages to be used.
    --
    -- After it was found that the general LUT overhead was so large that
    -- shift-by-multiplier does not constitute an improvement, the
    -- accompanying other changes were not implemented.
    --
    -- The comparisons regarding LUT usage would still remain valid,
    -- since the additional pipeline stages only use flip-flops.
    --
    -- The code has been left to allow test synthesis using the option.
    -- If a platform actually seems to benefit, the missing pipeline
    -- registers in dptc_output_packing could be implemented.

    use_mul: if (shift_mul = 1) generate
        mul_mask: for i in 0 to shift_mul_mask'left generate
            shift_mul_mask(i) <=
                '1' when (to_integer(shift_bits) = i) else '0';
        end generate;
        process (clk)
        begin
            if rising_edge(clk) then
                shift_input_reg <= shift_input;
                shift_mul_mask_reg <= shift_mul_mask;

                shifted_data_1 <=
                    std_logic_vector(unsigned(shift_input_reg) *
                                     unsigned(shift_mul_mask_reg));
                shifted_data_2 <= shifted_data_1;
                shifted_data_3 <= shifted_data_2;
                shifted_data <= shifted_data_3;
            end if;
        end process;
    end generate;
    
    -- Zero-extend the accumulated data-word.
    acc_word_ext <=
        (acc_word_ext'left downto acc_word'left+1=> '0') & acc_word;

    -- OR of the accumulated part with the new shifted word
    acc_out <= acc_word_ext or
               shifted_data(shifted_data'left downto ignore_low_bits);

    -- Update the accumulation buffer.  If the new accumulation buffer
    -- overflows, the 32 bits in the lowest part are marked valid as output,
    -- and then only the part in excess is kept.  Otherwise the new
    -- buffer (only the 31 bits low part) is kept.
    process (clk)
    begin
        if rising_edge(clk) then
            if (reset = '1') then
                acc_word <= (others => '0');
            elsif (new_word = '1') then
                acc_word <=
                    (acc_word'length -
                     (acc_out'length - out_word'length)-1 downto 0 => '0') &
                    acc_out(acc_out'left downto out_word'left+1);
            else
                acc_word <= acc_out(acc_word'range);
            end if;
        end if;
    end process;
    
    out_word <= acc_out(out_word'length-1 downto 0);
    
end RTL;
